def pytest_addoption(parser):
    """
    add `--show-viewer` as a valid command line flag
    """
    parser.addoption(
        "--show-viewer",
        action="store_true",
        default=False,
        help="don't show viewer during tests",
    )



import dask.array as da
import numpy as np
import pytest
import zarr
from numpy import ones, zeros

from ome_zarr.data import create_zarr
from ome_zarr.format import FormatV04
from ome_zarr.io import parse_url
from ome_zarr.reader import Node, Plate, Reader, Well
from ome_zarr.writer import (
    add_metadata,
    get_metadata,
    write_image,
    write_plate_metadata,
    write_well_metadata,
)


class TestReader:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        create_zarr(str(self.path))

    def assert_node(self, node: Node):
        if not node.data and not node.metadata:
            assert False, f"Empty node received: {node}"

    def test_image(self):
        reader = Reader(parse_url(str(self.path)))
        assert len(list(reader())) == 3

    def test_labels(self):
        filename = str(self.path.join("labels"))
        reader = Reader(parse_url(filename))
        assert len(list(reader())) == 3

    def test_label(self):
        filename = str(self.path.join("labels", "coins"))
        reader = Reader(parse_url(filename))
        assert len(list(reader())) == 3

    def test_omero(self):
        reader = Reader(parse_url(str(self.path)))()
        image_node = next(iter(reader))
        omero = image_node.zarr.root_attrs.get("omero")
        assert "channels" in omero
        assert isinstance(omero["channels"], list)
        assert len(omero["channels"]) == 1

    def test_read_v05(self):
        rng = np.random.default_rng(0)
        data = rng.poisson(lam=10, size=(10, 128, 128)).astype(np.uint8)
        img_path = str(self.path / "test_read_v05.zarr")
        root = zarr.group(img_path)
        arr = root.create_array(
            name="s0", shape=data.shape, chunks=(10, 10, 10), dtype=data.dtype
        )
        arr[:, :] = data
        root.attrs["ome"] = {
            "version": "0.5",
            "multiscales": [
                {
                    "datasets": [
                        {
                            "path": "s0",
                            "coordinateTransformations": [
                                {
                                    "type": "scale",
                                    "scale": [1, 1, 1],
                                }
                            ],
                        }
                    ]
                }
            ],
        }
        reader = Reader(parse_url(img_path))
        nodes = list(reader())
        assert len(nodes) == 1
        image_node = nodes[0]
        assert np.allclose(data, image_node.data[0])


class TestInvalid:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")

    def test_invalid_version(self):
        grp = create_zarr(str(self.path))
        # update version to something invalid
        attrs = get_metadata(grp)
        attrs["multiscales"][0]["version"] = "invalid"
        add_metadata(grp, attrs)
        # should raise exception
        with pytest.raises(ValueError) as exe:
            reader = Reader(parse_url(str(self.path)))
            assert len(list(reader())) == 2
        assert str(exe.value) == "Version invalid not recognized"


class TestHCSReader:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        self.store = parse_url(str(self.path), mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)

    def test_minimal_plate(self):
        write_plate_metadata(self.root, ["A"], ["1"], ["A/1"])
        row_group = self.root.require_group("A")
        well = row_group.require_group("1")
        write_well_metadata(well, ["0"])
        image = well.require_group("0")
        write_image(zeros((1, 1, 1, 256, 256)), image)

        reader = Reader(parse_url(str(self.path)))
        nodes = list(reader())
        # currently reading plate labels disabled. Only 1 node
        assert len(nodes) == 1
        assert len(nodes[0].specs) == 1
        assert isinstance(nodes[0].specs[0], Plate)
        # assert len(nodes[1].specs) == 1
        # assert isinstance(nodes[1].specs[0], PlateLabels)

    @pytest.mark.parametrize("field_paths", [["0", "1", "2"], ["img1", "img2", "img3"]])
    def test_multiwells_plate(self, field_paths):
        row_names = ["A", "B", "C"]
        col_names = ["1", "2", "3", "4"]
        well_paths = ["A/1", "A/2", "A/4", "B/2", "B/3", "C/1", "C/3", "C/4"]
        write_plate_metadata(self.root, row_names, col_names, well_paths)
        for wp in well_paths:
            row, col = wp.split("/")
            row_group = self.root.require_group(row)
            well = row_group.require_group(col)
            write_well_metadata(well, field_paths)
            for field in field_paths:
                image = well.require_group(str(field))
                write_image(ones((1, 1, 1, 256, 256)), image)

        reader = Reader(parse_url(str(self.path)))
        nodes = list(reader())
        # currently reading plate labels disabled. Only 1 node
        assert len(nodes) == 1

        plate_node = nodes[0]
        assert len(plate_node.specs) == 1
        assert isinstance(plate_node.specs[0], Plate)
        # data should be a Dask array
        pyramid = plate_node.data
        assert isinstance(pyramid[0], da.Array)
        # if we compute(), expect to get numpy array
        result = pyramid[0].compute()
        assert isinstance(result, np.ndarray)

        # Get the plate node's array. It should be fused from the first field of all
        # well arrays (which in this test are non-zero), with zero values for wells
        # that failed to load (not expected) or the surplus area not filled by a well.
        expected_num_pixels = (
            len(well_paths) * len(field_paths[:1]) * np.prod((1, 1, 1, 256, 256))
        )
        pyramid_0 = pyramid[0]
        assert np.asarray(pyramid_0).sum() == expected_num_pixels

        # assert isinstance(nodes[1].specs[0], PlateLabels)

        reader = Reader(parse_url(f"{self.path}/{well_paths[0]}"))
        nodes = list(reader())
        assert isinstance(nodes[0].specs[0], Well)
        pyramid = nodes[0].data
        assert isinstance(pyramid[0], da.Array)
        result = pyramid[0].compute()
        assert isinstance(result, np.ndarray)



import pytest
import zarr
from numpy import zeros

from ome_zarr.data import create_zarr
from ome_zarr.format import FormatV01, FormatV02, FormatV03, FormatV04
from ome_zarr.io import parse_url
from ome_zarr.reader import Label, Labels, Multiscales, Node, Plate, Well
from ome_zarr.writer import write_image, write_plate_metadata, write_well_metadata


class TestNode:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        create_zarr(str(self.path))

    def test_image(self):
        node = Node(parse_url(str(self.path)), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 2
        assert isinstance(node.specs[0], Multiscales)

    def test_labels(self):
        filename = str(self.path.join("labels"))
        node = Node(parse_url(filename), list())
        assert not node.data
        assert not node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Labels)

    def test_label(self):
        filename = str(self.path.join("labels", "coins"))
        node = Node(parse_url(filename), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 2
        assert isinstance(node.specs[0], Label)
        assert isinstance(node.specs[1], Multiscales)


class TestHCSNode:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        self.store = parse_url(str(self.path), mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)

    def test_minimal_plate(self):
        write_plate_metadata(self.root, ["A"], ["1"], ["A/1"], fmt=FormatV01())
        row_group = self.root.require_group("A")
        well = row_group.require_group("1")
        write_well_metadata(well, ["0"], fmt=FormatV04())
        image = well.require_group("0")
        write_image(zeros((1, 1, 1, 256, 256)), image, fmt=FormatV01())

        node = Node(parse_url(str(self.path)), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Plate)
        assert node.specs[0].row_names == ["A"]
        assert node.specs[0].col_names == ["1"]
        assert node.specs[0].well_paths == ["A/1"]
        assert node.specs[0].row_count == 1
        assert node.specs[0].column_count == 1

        node = Node(parse_url(str(self.path / "A" / "1")), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Well)

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02(), FormatV03()))
    def test_multiwells_plate(self, fmt):
        row_names = ["A", "B", "C"]
        col_names = ["1", "2", "3", "4"]
        well_paths = ["A/1", "A/2", "A/4", "B/2", "B/3", "C/1", "C/3", "C/4"]
        write_plate_metadata(self.root, row_names, col_names, well_paths, fmt=fmt)
        for wp in well_paths:
            row, col = wp.split("/")
            row_group = self.root.require_group(row)
            well = row_group.require_group(col)
            write_well_metadata(well, ["0", "1", "2"], fmt=fmt)
            for field in range(3):
                image = well.require_group(str(field))
                write_image(zeros((1, 1, 1, 256, 256)), image, fmt=fmt)

        node = Node(parse_url(str(self.path)), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Plate)
        assert node.specs[0].row_names == row_names
        assert node.specs[0].col_names == col_names
        assert node.specs[0].well_paths == well_paths
        assert node.specs[0].row_count == 3
        assert node.specs[0].column_count == 4

        for wp in well_paths:
            node = Node(parse_url(str(self.path / wp)), list())
            assert node.data
            assert node.metadata
            assert len(node.specs) == 1
            assert isinstance(node.specs[0], Well)

        empty_wells = ["A/3", "B/1", "B/4", "C/2"]
        for wp in empty_wells:
            assert parse_url(str(self.path / wp)) is None

    @pytest.mark.parametrize(
        "axes, dims",
        (
            (["y", "x"], (256, 256)),
            (["t", "y", "x"], (1, 256, 256)),
            (["z", "y", "x"], (1, 256, 256)),
            (["c", "y", "x"], (1, 256, 256)),
            (["c", "z", "y", "x"], (1, 1, 256, 256)),
            (["t", "z", "y", "x"], (1, 1, 256, 256)),
            (["t", "c", "y", "x"], (1, 1, 256, 256)),
        ),
    )
    def test_plate_2D5D(self, axes, dims):
        write_plate_metadata(self.root, ["A"], ["1"], ["A/1"], fmt=FormatV03())
        row_group = self.root.require_group("A")
        well = row_group.require_group("1")
        write_well_metadata(well, ["0"], fmt=FormatV03())
        image = well.require_group("0")
        write_image(zeros(dims), image, fmt=FormatV03(), axes=axes)

        node = Node(parse_url(str(self.path)), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Plate)

        node = Node(parse_url(str(self.path / "A" / "1")), list())
        assert node.data
        assert node.metadata
        assert len(node.specs) == 1
        assert isinstance(node.specs[0], Well)



"""Utility methods for ome_zarr access."""

import csv
import json
import logging
import os
import urllib
import webbrowser
import xml.etree.ElementTree as ET
from collections.abc import Iterator
from datetime import datetime
from http.server import (  # type: ignore[attr-defined]
    HTTPServer,
    SimpleHTTPRequestHandler,
    test,
)
from pathlib import Path

import dask
import dask.array as da
import zarr
from dask.diagnostics import ProgressBar

from .format import format_from_version
from .io import parse_url
from .reader import Multiscales, Node, Reader
from .types import Any, JSONDict

LOGGER = logging.getLogger("ome_zarr.utils")


def info(path: str, stats: bool = False) -> Iterator[Node]:
    """Print information about an OME-Zarr fileset.

    All :class:`Nodes <ome_utils.reader.Node>` that are found from the given path will
    be visited recursively.

    Parameters
    ----------
    path :
        Path to OME-Zarr fileset.
    stats :
        If True, print stats (currently just minimum/maximum of all arrays)

    Warnings
    --------
    Passing ``stats=True`` will trigger a full read of every array in the fileset.
    """
    zarr = parse_url(path)
    assert zarr, f"not a zarr: {zarr}"
    reader = Reader(zarr)
    for node in reader():
        if not node.specs:
            print(f"not an ome-zarr node: {node}")
            continue

        print(node)
        loc = node.zarr
        version = loc.zgroup.get("version")
        if version is None:
            version = loc.zgroup.get("multiscales", [{}])[0].get("version", "")
        print(" - version:", version)
        print(" - metadata")
        for spec in node.specs:
            print(f"   - {spec.__class__.__name__}")
        print(" - data")
        for array in node.data:
            minmax = ""
            if stats:
                minmax = f" minmax={dask.compute(array.min(), array.max())}"
            print(f"   - {array.shape}{minmax}")
        LOGGER.debug(node.data)
        yield node


def view(input_path: str, port: int = 8000, dry_run: bool = False) -> None:
    # serve the parent directory in a simple server with CORS. Open browser
    # dry_run is for testing, so we don't open the browser or start the server

    zarrs = []
    if (Path(input_path) / ".zattrs").exists() or (
        Path(input_path) / "zarr.json"
    ).exists():
        zarrs = find_multiscales(Path(input_path))
    if len(zarrs) == 0:
        print(
            f"No OME-Zarr images found in {input_path}. "
            f"Try $ ome_zarr finder {input_path}"
        )
        return

    parent_dir, image_name = os.path.split(input_path)
    if len(image_name) == 0:
        parent_dir, image_name = os.path.split(parent_dir)
    parent_dir = str(parent_dir)

    # open ome-ngff-validator in a web browser...
    url = (
        f"https://ome.github.io/ome-ngff-validator/"
        f"?source=http://localhost:{port}/{image_name}"
    )

    class CORSRequestHandler(SimpleHTTPRequestHandler):
        def end_headers(self) -> None:
            self.send_header("Access-Control-Allow-Origin", "*")
            SimpleHTTPRequestHandler.end_headers(self)

        def translate_path(self, path: str) -> str:
            # Since we don't call the class constructor ourselves,
            # we set the directory here instead
            self.directory = parent_dir
            super_path = super().translate_path(path)
            return super_path

    # for testing
    if dry_run:
        return

    # Open in browser...
    webbrowser.open(url)

    # ...then start serving content
    test(CORSRequestHandler, HTTPServer, port=port)


def find_multiscales(path_to_zattrs):
    # return list of images. Each image is [path_to_zarr, name, dirname]
    # We want full path to find the multiscales Image. e.g. full/path/to/image.zarr/0
    # AND we want image Name, e.g. "image.zarr Series 0"
    # AND we want the dir path to use for Tags e.g. full/path/to
    text = None
    for name in (".zattrs", "zarr.json"):
        if (Path(path_to_zattrs) / name).exists():
            with open(path_to_zattrs / name) as f:
                text = f.read()
            break
    if text is None:
        print("No .zattrs or zarr.json found in {path_to_zattrs}")
        return []
    zattrs = json.loads(text)
    if "attributes" in zattrs and "ome" in zattrs["attributes"]:
        zattrs = zattrs["attributes"]["ome"]
    if "plate" in zattrs:
        plate = zattrs.get("plate")
        wells = plate.get("wells")
        field = "0"
        if len(wells) > 0:
            path_to_zarr = path_to_zattrs / wells[0].get("path") / field
            plate_name = os.path.basename(path_to_zattrs)
            return [[path_to_zarr, plate_name, os.path.dirname(path_to_zattrs)]]
        else:
            LOGGER.info("No wells found in plate%s", path_to_zattrs)
            return []
    elif zattrs.get("bioformats2raw.layout") == 3:
        # Open OME/METADATA.ome.xml
        try:
            tree = ET.parse(path_to_zattrs / "OME" / "METADATA.ome.xml")
            root = tree.getroot()
            # spec says "If the "series" attribute does not exist and no "plate" is
            # present, separate "multiscales" images MUST be stored in consecutively
            # numbered groups starting from 0 (i.e. "0/", "1/", "2/", "3/", ...)."
            series = 0
            images = []
            for child in root:
                # tag is eg. {http://www.openmicroscopy.org/Schemas/OME/2016-06}Image
                if child.tag.endswith("Image"):
                    img_name = f"{os.path.basename(path_to_zattrs)} Series:{series}"
                    # Get Name from XML metadata, otherwise use path and Series
                    img_name = child.attrib.get("Name", img_name)
                    images.append(
                        [
                            path_to_zattrs / str(series),
                            img_name,
                            os.path.dirname(path_to_zattrs),
                        ]
                    )
                    series += 1
            return images
        except Exception as ex:  # noqa: BLE001
            print(ex)
    elif zattrs.get("multiscales"):
        return [
            [
                path_to_zattrs,
                os.path.basename(path_to_zattrs),
                os.path.dirname(path_to_zattrs),
            ]
        ]
    return []


def splitall(path):
    # Use os.path.split() repeatedly to split path into dirs
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def finder(input_path: str, port: int = 8000, dry_run=False) -> None:
    # serve the parent directory in a simple server with CORS. Open browser
    # dry_run is for testing, so we don't open the browser or start the server
    parent_path, server_dir = os.path.split(input_path)
    # in case input_path had trailing slash, we go one level up...
    if len(server_dir) == 0:
        parent_path, server_dir = os.path.split(parent_path)

    # 'input_path' is path passed to the script. To the data dir. E.g. "ZARR/data"
    # 'parent_path', e.g. "ZARR" just for running http server
    # 'server_dir' is the name of our top-level dir E.g. "data"

    # We will be serving the data from last dir in /parent/dir/path
    # so we need to use that as base for image URLs...

    # walk the input path to find all .zattrs files...
    def walk(path: Path):
        if (path / ".zattrs").exists() or (path / "zarr.json").exists():
            yield from find_multiscales(path)
        else:
            for p in path.iterdir():
                if (p / ".zattrs").exists() or (p / "zarr.json").exists():
                    yield from find_multiscales(p)
                elif p.is_dir():
                    yield from walk(p)
                else:
                    continue

    url = None
    zarrs = list(walk(Path(input_path)))

    # If we have just one zarr, open ome-ngff-validator in a web browser...
    if len(zarrs) == 0:
        print("No OME-Zarr files found in", input_path)
        return
    else:
        # ...otherwise write to CSV file and open in BioFile Finder
        col_names = ["File Path", "File Name", "Folders", "Uploaded"]
        # write csv file into the dir we're serving from...
        bff_csv = os.path.join(input_path, "biofile_finder.csv")

        with open(bff_csv, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            writer.writerow(col_names)
            for zarr_img in zarrs:
                # zarr paths start with full path to img
                # e.g. ZARR/data/to/img (from walk("ZARR/data"))
                # but we want them to be from the server_dir to img, e.g "data/to/img".
                # So we want relative /to/img path, from input_path -> to img
                relpath = os.path.relpath(zarr_img[0], input_path)
                # On Windows, we need to replace \\ with / in relpath for URL
                rel_url = "/".join(splitall(relpath))
                file_path = f"http://localhost:{port}/{server_dir}/{rel_url}"
                name = zarr_img[1] or os.path.basename(zarr_img[0])
                # folders is "f1,f2,f3" etc.
                folders_path = os.path.relpath(zarr_img[2], input_path)
                folders = ",".join(splitall(folders_path))
                timestamp = ""
                try:
                    mtime = os.path.getmtime(zarr_img[0])
                    # format mtime as "YYYY-MM-DD HH:MM:SS.Z"
                    timestamp = datetime.fromtimestamp(mtime).strftime(
                        "%Y-%m-%d %H:%M:%S.%Z"
                    )
                except OSError:
                    pass
                writer.writerow([file_path, name, folders, timestamp])

        source = {
            "uri": f"http://localhost:{port}/{server_dir}/biofile_finder.csv",
            "type": "csv",
            "name": "biofile_finder.csv",
        }
        s = urllib.parse.quote(json.dumps(source))
        url = f"https://bff.allencell.org/app?source={s}"
        # show small thumbnails view by default. (v=3 for big thumbnails)
        url += "&v=2"

    class CORSRequestHandler(SimpleHTTPRequestHandler):
        def end_headers(self) -> None:
            self.send_header("Access-Control-Allow-Origin", "*")
            SimpleHTTPRequestHandler.end_headers(self)

        def translate_path(self, path: str) -> str:
            # Since we don't call the class constructor ourselves,
            # we set the directory here instead
            self.directory = parent_path
            super_path = super().translate_path(path)
            return super_path

    # for testing
    if dry_run:
        return

    # Open in browser...
    webbrowser.open(url)

    # ...then start serving content
    test(CORSRequestHandler, HTTPServer, port=port)


def download(input_path: str, output_dir: str = ".") -> None:
    """Download an OME-Zarr from the given path.

    All :class:`Nodes <ome_utils.reader.Node>` that are found from the given path will
    be included in the download.
    """
    location = parse_url(input_path)
    assert location, f"not a zarr: {location}"

    reader = Reader(location)
    nodes: list[Node] = list()
    paths: list[list[str]] = list()
    for node in reader():
        nodes.append(node)
        paths.append(node.zarr.parts())

    common = strip_common_prefix(paths)
    output_path = Path(output_dir)
    root_path = output_path / common

    assert not root_path.exists(), f"{root_path} already exists!"
    print("downloading...")
    for path in paths:
        print("  ", Path(*path))
    print(f"to {output_dir}")

    for path, node in sorted(zip(paths, nodes)):
        target_path = output_path / Path(*path)
        target_path.mkdir(parents=True)

        # Use version etc...
        version = node.zarr.version
        fmt = format_from_version(version)

        # store = parse_url(input_path, mode="w", fmt=fmt)
        group_file = "zarr.json"
        attrs_file = "zarr.json"
        if fmt.zarr_format == 2:
            group_file = ".zgroup"
            attrs_file = ".zattrs"

        with (target_path / group_file).open("w") as f:
            f.write(json.dumps(node.zarr.zgroup))
        with (target_path / attrs_file).open("w") as f:
            metadata: JSONDict = {}
            node.write_metadata(metadata)
            if fmt.zarr_format == 3:
                # For zarr v3, we need to put metadata under "ome" namespace
                metadata = {
                    "attributes": {"ome": metadata},
                    "zarr_format": 3,
                    "node_type": "group",
                }
            f.write(json.dumps(metadata))

        resolutions: list[da.core.Array] = []
        datasets: list[str] = []

        for spec in node.specs:
            if isinstance(spec, Multiscales):
                datasets = spec.datasets
                resolutions = node.data
                options: dict[str, Any] = {}
                if fmt.zarr_format == 2:
                    options["dimension_separator"] = "/"
                else:
                    options["chunk_key_encoding"] = fmt.chunk_key_encoding
                    options["dimension_names"] = [
                        axis["name"] for axis in node.metadata["axes"]
                    ]
                if datasets and resolutions:
                    pbar = ProgressBar()
                    for dataset, data in reversed(list(zip(datasets, resolutions))):
                        LOGGER.info("resolution %s...", dataset)
                        with pbar:
                            data.to_zarr(
                                str(target_path / dataset),
                                zarr_format=fmt.zarr_format,
                                **options,
                            )
            else:
                # Assume a group that needs metadata, like labels
                zarr.group(str(target_path))


def strip_common_prefix(parts: list[list[str]]) -> str:
    """Find and remove the prefix common to all strings.

    Returns the last element of the common prefix.
    An exception is thrown if no common prefix exists.

    >>> paths = [["a", "b"], ["a", "b", "c"]]
    >>> strip_common_prefix(paths)
    'b'
    >>> paths
    [['b'], ['b', 'c']]
    """
    first_mismatch = 0
    min_length = min(len(x) for x in parts)

    for idx in range(min_length):
        if len({x[idx] for x in parts}) == 1:
            first_mismatch += 1
        else:
            break

    if first_mismatch <= 0:
        msg = "No common prefix:\n"
        for path in parts:
            msg += f"{path}\n"
        raise Exception(msg)
    else:
        common = parts[0][first_mismatch - 1]

    for idx, path in enumerate(parts):
        parts[idx] = parts[idx][first_mismatch - 1 :]

    return common



import pathlib

import numpy as np
import pytest
import zarr

from ome_zarr.format import CurrentFormat
from ome_zarr.format import FormatV01 as V01
from ome_zarr.format import FormatV02 as V02
from ome_zarr.format import FormatV03 as V03
from ome_zarr.io import parse_url
from ome_zarr.reader import Multiscales, Reader
from ome_zarr.writer import write_image


class TestUpgrade:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))

    def create_data(self, shape, fmt=CurrentFormat(), dtype=np.uint8, mean_val=128):
        self.store = parse_url(self.path, mode="w", fmt=fmt).store
        self.root = zarr.group(store=self.store)
        self.group = self.root.create_group("test")
        rng = np.random.default_rng(0)
        return rng.poisson(mean_val, size=shape).astype(dtype)

    def assert_data(self, path, shape, fmt, mode="r"):
        loc = parse_url(path, mode=mode, fmt=fmt)
        assert loc
        reader = Reader(loc)
        node = next(iter(reader()))
        assert Multiscales.matches(node.zarr)
        assert node.data[0].shape == shape
        assert np.max(node.data[0]) > 0
        assert loc.fmt == fmt

    @pytest.mark.parametrize(
        "path, version",
        (("v1", V01()), ("v2", V02())),
    )
    def test_pre_created(self, request, path, version):
        shape = (1, 2, 1, 16, 16)
        self.assert_data(f"{request.fspath.dirname}/data/{path}", shape, version)

    @pytest.mark.parametrize(
        "version",
        (
            pytest.param(V01(), id="V01"),
            pytest.param(V02(), id="V02"),
            pytest.param(V03(), id="V03"),
        ),
    )
    def test_newly_created(self, version):
        shape = (1, 1, 1, 8, 8)
        data = self.create_data(shape, version)
        axes = None
        if version not in ("0.1", "0.2"):
            axes = "tczyx"
        write_image(image=data, group=self.group, scaler=None, fmt=version, axes=axes)
        self.assert_data(f"{self.path}/test", shape=shape, fmt=version)

    def test_requested_no_upgrade(self):
        print("warnings")

    def test_automatic_upgrade(self):
        print("info or debug?")

    def test_cli_upgrade(self):
        print("dry-run?")



import pytest

from ome_zarr.data import create_zarr
from ome_zarr.io import parse_url
from ome_zarr.reader import OMERO, Label, Labels, Multiscales, Node, Spec


class TestStartingPoints:
    """Creates a small but complete OME-Zarr file and tests that readers will detect the
    correct type when starting at all the various levels."""

    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        create_zarr(str(self.path))

    def matches(self, node: Node, expected: list[type[Spec]]):
        found: list[type[Spec]] = [type(spec) for spec in node.specs]

        expected_names = sorted(x.__name__ for x in expected)
        found_names = sorted(x.__name__ for x in found)
        assert expected_names == found_names

    def get_spec(self, node: Node, spec_type: type[Spec]):
        for spec in node.specs:
            if isinstance(spec, spec_type):
                return spec
        assert False, f"no {spec_type} found"

    def test_top_level(self):
        zarr = parse_url(str(self.path))
        assert zarr is not None
        node = Node(zarr, list())
        self.matches(node, {Multiscales, OMERO})
        multiscales = self.get_spec(node, Multiscales)
        assert multiscales.lookup("multiscales", [])

    def test_labels(self):
        zarr = parse_url(str(self.path + "/labels"))
        assert zarr is not None
        node = Node(zarr, list())
        self.matches(node, {Labels})

    def test_label(self):
        zarr = parse_url(str(self.path + "/labels/coins"))
        assert zarr is not None
        node = Node(zarr, list())
        self.matches(node, {Label, Multiscales})
        multiscales = self.get_spec(node, Multiscales)
        assert multiscales.lookup("multiscales", [])



from pathlib import Path

import pytest
import zarr
from zarr.storage import LocalStore

from ome_zarr.data import create_zarr
from ome_zarr.io import ZarrLocation, parse_url
from ome_zarr.writer import add_metadata, get_metadata


class TestIO:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        create_zarr(str(self.path))
        self.store = parse_url(str(self.path), mode="r").store
        self.root = zarr.open_group(store=self.store, mode="r")

    def test_parse_url(self):
        assert parse_url(str(self.path))

    def test_parse_nonexistent_url(self):
        assert parse_url(str(self.path + "/does-not-exist")) is None

    def test_loc_str(self):
        assert ZarrLocation(str(self.path))

    def test_loc_path(self):
        assert ZarrLocation(Path(self.path))

    def test_loc_store(self):
        assert ZarrLocation(self.store)

    def test_loc_fs(self):
        store = LocalStore(str(self.path))
        loc = ZarrLocation(store)
        assert loc

    def test_no_overwrite(self):
        print("self.path:", self.path)

        assert self.root.attrs.get("ome") is not None
        # Test that we can open a store to write, without
        # overwriting existing data
        new_store = parse_url(str(self.path), mode="w").store
        new_root = zarr.open_group(store=new_store)
        add_metadata(new_root, {"extra": "test_no_overwrite"})
        # read...
        read_store = parse_url(str(self.path)).store
        read_root = zarr.open_group(store=read_store, mode="r")
        attrs = get_metadata(read_root)
        assert attrs.get("extra") == "test_no_overwrite"
        assert attrs.get("multiscales") is not None



"""Definition of complex types for use elsewhere."""

from collections.abc import Callable
from typing import Any

LayerData = tuple[Any] | tuple[Any, dict] | tuple[Any, dict, str]

PathLike = str | list[str]

ReaderFunction = Callable[[PathLike], list[LayerData]]

JSONDict = dict[str, Any]



import pathlib

import dask.array as da
import numpy as np
import pytest
import zarr

from ome_zarr.io import parse_url
from ome_zarr.scale import Scaler
from ome_zarr.writer import write_image


class TestScaler:
    @pytest.fixture(
        params=(
            (1, 2, 1, 256, 256),
            (3, 512, 512),
            (256, 256),
        ),
        ids=["5D", "3D", "2D"],
    )
    def shape(self, request):
        return request.param

    def create_data(self, shape, dtype=np.uint8, mean_val=10):
        rng = np.random.default_rng(0)
        return rng.poisson(mean_val, size=shape).astype(dtype)

    def check_downscaled(self, downscaled, shape, scale_factor=2):
        expected_shape = shape
        for data in downscaled:
            assert data.shape == expected_shape
            assert data.dtype == downscaled[0].dtype
            expected_shape = expected_shape[:-2] + tuple(
                sh // scale_factor for sh in expected_shape[-2:]
            )

    def test_nearest(self, shape):
        data = self.create_data(shape)
        scaler = Scaler()
        downscaled = scaler.nearest(data)
        self.check_downscaled(downscaled, shape)

    def test_nearest_via_method(self, shape):
        data = self.create_data(shape)

        scaler = Scaler()
        expected_downscaled = scaler.nearest(data)

        scaler.method = "nearest"
        downscaled = scaler.func(data)
        self.check_downscaled(downscaled, shape)

        assert (
            np.sum(
                [
                    not np.array_equal(downscaled[i], expected_downscaled[i])
                    for i in range(len(downscaled))
                ]
            )
            == 0
        )

    # this fails because of wrong channel dimension; need to fix in follow-up PR
    @pytest.mark.xfail
    def test_gaussian(self, shape):
        data = self.create_data(shape)
        scaler = Scaler()
        downscaled = scaler.gaussian(data)
        self.check_downscaled(downscaled, shape)

    # this fails because of wrong channel dimension; need to fix in follow-up PR
    @pytest.mark.xfail
    def test_laplacian(self, shape):
        data = self.create_data(shape)
        scaler = Scaler()
        downscaled = scaler.laplacian(data)
        self.check_downscaled(downscaled, shape)

    def test_local_mean(self, shape):
        data = self.create_data(shape)
        scaler = Scaler()
        downscaled = scaler.local_mean(data)
        self.check_downscaled(downscaled, shape)

    def test_local_mean_via_method(self, shape):
        data = self.create_data(shape)

        scaler = Scaler()
        expected_downscaled = scaler.local_mean(data)

        scaler.method = "local_mean"
        downscaled = scaler.func(data)
        self.check_downscaled(downscaled, shape)

        assert (
            np.sum(
                [
                    not np.array_equal(downscaled[i], expected_downscaled[i])
                    for i in range(len(downscaled))
                ]
            )
            == 0
        )

    @pytest.mark.skip(reason="This test does not terminate")
    def test_zoom(self, shape):
        data = self.create_data(shape)
        scaler = Scaler()
        downscaled = scaler.zoom(data)
        self.check_downscaled(downscaled, shape)

    def test_scale_dask(self, shape):
        data = self.create_data(shape)
        # chunk size gives odd-shaped chunks at the edges
        # tests https://github.com/ome/ome-zarr-py/pull/244
        chunk_size = [100, 100]
        chunk_2d = (*(1,) * (data.ndim - 2), *chunk_size)

        data_delayed = da.from_array(data, chunks=chunk_2d)

        scaler = Scaler()
        resized_data = scaler.resize_image(data)
        resized_dask = scaler.resize_image(data_delayed)

        assert np.array_equal(resized_data, resized_dask)

    def test_scale_dask_via_method(self, shape):
        data = self.create_data(shape)

        chunk_size = [100, 100]
        chunk_2d = (*(1,) * (data.ndim - 2), *chunk_size)
        data_delayed = da.from_array(data, chunks=chunk_2d)

        scaler = Scaler()
        expected_downscaled = scaler.resize_image(data)

        scaler.method = "resize_image"
        assert np.array_equal(expected_downscaled, scaler.func(data))
        assert np.array_equal(expected_downscaled, scaler.func(data_delayed))

    def test_big_dask_pyramid(self, tmpdir):
        # from https://github.com/ome/omero-cli-zarr/pull/134
        shape = (6675, 9560)
        data = self.create_data(shape)
        data_delayed = da.from_array(data, chunks=(1000, 1000))
        print("data_delayed", data_delayed)
        scaler = Scaler()
        level_1 = scaler.resize_image(data_delayed)
        print("level_1", level_1)
        # to zarr invokes compute
        data_dir = tmpdir.mkdir("test_big_dask_pyramid")
        da.to_zarr(level_1, str(data_dir))

    @pytest.mark.parametrize("method", ["gaussian", "laplacian"])
    def test_pyramid_args(self, shape, tmpdir, method):
        path = pathlib.Path(tmpdir.mkdir("data"))
        store = parse_url(path, mode="w").store
        group = zarr.group(store=store, overwrite=True)
        data = self.create_data(shape)

        scaler = Scaler(
            downscale=2,
            method=method,
            max_layer=2,
        )

        axes = "tczyx"[-len(shape) :]
        write_image(
            image=data,
            group=group,
            scaler=scaler,
            axes=axes,
        )



import logging
import os

import pytest

from ome_zarr.data import astronaut, create_zarr
from ome_zarr.utils import download, info


def log_strings(idx, c, y, x, cc, cy, cx, dtype):
    yield f"resolution: {idx}"
    yield f" - shape ('c', 'y', 'x') = ({c}, {y}, {x})"
    yield f" - chunks =  ['{cc}', '{cx}', '{cy}']"
    yield f" - dtype = {dtype}"


class TestOmeZarr:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = tmpdir.mkdir("data")
        create_zarr(str(self.path), method=astronaut)

    def check_info_stdout(self, out):
        for log in log_strings(0, 3, 1024, 1024, 1, 64, 64, "uint8"):
            assert log in out
        for log in log_strings(1, 3, 512, 512, 1, 64, 64, "int8"):
            assert log in out

        # from info's print of omero metadata
        # note: some metadata is no longer handled by info but rather
        #       in the ome_zarr.napari.transform method

    def test_info(self, caplog):
        with caplog.at_level(logging.DEBUG):
            list(info(str(self.path)))
        self.check_info_stdout(caplog.text)

    def test_download(self, capsys, caplog, tmpdir):
        target = str(tmpdir / "out")
        with caplog.at_level(logging.DEBUG):
            download(str(self.path), output_dir=target)
            download_zarr = os.path.join(target, "data")
            assert os.path.exists(download_zarr)
            info(download_zarr)
        self.check_info_stdout(caplog.text)
        # check download progress in stdout
        out = capsys.readouterr().out
        assert "100% Completed" in out



import pathlib
import sys

# alternative is to make code installable (which it is!)
sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())


extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
]

# use index.rst instead of contents.rst
master_doc = "index"

# -- Project information -----------------------------------------------------

project = "ome-zarr-py"
copyright = "Open Microscopy Environment"  # noqa: A001
author = "Open Microscopy Environment"

# Example configuration for intersphinx: refer to the Python standard library.
# use in refs e.g:
# :ref:`comparison manual <python:comparisons>`
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "zarr": ("https://zarr.readthedocs.io/en/stable/", None),
    "dask": ("https://docs.dask.org/en/stable", None),
}

# https://github.com/readthedocs/sphinx_rtd_theme
html_theme = "sphinx_rtd_theme"



import os
from collections import deque
from pathlib import Path

import pytest
import zarr

from ome_zarr.cli import main
from ome_zarr.format import CurrentFormat, FormatV04, FormatV05
from ome_zarr.io import parse_url
from ome_zarr.utils import find_multiscales, finder, strip_common_prefix, view
from ome_zarr.writer import write_plate_metadata


def directory_items(directory: Path):
    """
    Get all items (files and folders) in a directory, relative to that directory.
    """
    if not directory.is_dir():
        raise ValueError(f"{directory} is not a directory")

    return sorted([p.relative_to(directory) for p in directory.glob("*")])


class TestCli:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = (tmpdir / "data").mkdir()

    @pytest.fixture(params=["0.1", "0.2", "0.3"], ids=["v0.1", "v0.2", "v0.3"])
    def s3_address(self, request):
        urls = {
            "0.1": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.1/6001240.zarr",
            "0.2": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.2/6001240.zarr",
            "0.3": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.3/9836842.zarr",
            "0.4": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0062A/6001240.zarr",
            "0.5": "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0062A/6001240_labels.zarr",
        }
        return urls[request.param]

    @pytest.mark.parametrize(
        "fmt",
        (
            pytest.param(FormatV04(), id="V04"),
            pytest.param(FormatV05(), id="V05"),
            pytest.param(None, id="CurrentFormat"),
        ),
    )
    def test_coins_info(self, capsys, fmt):
        """Test create and info with various formats."""
        filename = f"{self.path}-1"
        args = ["create", "--method=coins", filename]
        if fmt:
            args += ["--format", fmt.version]
        main(args)
        main(["info", filename])
        out, err = capsys.readouterr()
        print("Captured output:", out)
        assert os.path.join("labels", "coins") in out
        version = fmt.version if fmt else CurrentFormat().version
        assert f"- version: {version}" in out

    def test_astronaut_info(self):
        filename = f"{self.path}-2"
        main(["create", "--method=astronaut", filename])
        main(["info", filename])

    @pytest.mark.parametrize(
        "fmt",
        (
            pytest.param(FormatV04(), id="V04"),
            pytest.param(FormatV05(), id="V05"),
            pytest.param(None, id="CurrentFormat"),
        ),
    )
    def test_astronaut_download(self, tmpdir, fmt):
        out = str(tmpdir / "out")
        filename = f"{self.path}-3"
        basename = os.path.split(filename)[-1]
        args = ["create", "--method=astronaut", filename]
        if fmt:
            args += ["--format", fmt.version]
        main(args)
        main(["download", filename, f"--output={out}"])
        main(["info", f"{out}/{basename}"])

        if fmt is not None and fmt.zarr_format == 2:
            assert directory_items(Path(out) / "data-3") == [
                Path(".zattrs"),
                Path(".zgroup"),
                Path("0"),
                Path("1"),
                Path("2"),
                Path("3"),
                Path("4"),
                Path("labels"),
            ]
            assert directory_items(Path(out) / "data-3" / "1") == [
                Path(".zarray"),
                Path(".zattrs"),  # empty '{}'
                Path("0"),
                Path("1"),
                Path("2"),
            ]
        else:
            assert directory_items(Path(out) / "data-3") == [
                Path("0"),
                Path("1"),
                Path("2"),
                Path("3"),
                Path("4"),
                Path("labels"),
                Path("zarr.json"),
            ]
            assert directory_items(Path(out) / "data-3" / "1") == [
                Path("c"),
                Path("zarr.json"),
            ]

    def test_s3_info(self, s3_address):
        main(["info", s3_address])

    def test_strip_prefix_relative(self):
        top = Path(".") / "d"
        mid = Path(".") / "d" / "e"
        bot = Path(".") / "d" / "e" / "f"
        self._rotate_and_test(top, mid, bot)

    def test_strip_prefix_absolute(self):
        top = Path("/") / "a" / "b" / "c" / "d"
        mid = Path("/") / "a" / "b" / "c" / "d" / "e"
        bot = Path("/") / "a" / "b" / "c" / "d" / "e" / "f"
        self._rotate_and_test(top, mid, bot)

    def _rotate_and_test(self, *hierarchy: Path, reverse: bool = True):
        results: list[list[str]] = [
            list((Path("d")).parts),
            list((Path("d") / "e").parts),
            list((Path("d") / "e" / "f").parts),
        ]
        for x in range(3):
            firstpass = deque(hierarchy)
            firstpass.rotate(1)

            copy = [list(x.parts) for x in firstpass]
            common = strip_common_prefix(copy)
            assert common == "d"
            assert {tuple(x) for x in copy} == {tuple(x) for x in results}

        if reverse:
            secondpass: deque = deque(hierarchy)
            secondpass.reverse()
            self._rotate_and_test(*list(secondpass), reverse=False)

    def test_view(self):
        filename = f"{self.path}-4"
        main(["create", "--method=astronaut", filename])
        # CLI doesn't support the dry_run option yet
        # main(["view", filename, "8000"])
        # we need dry_run to be True to avoid blocking the test with server
        view(filename, 8000, True)

    @pytest.mark.parametrize(
        "fmt",
        (pytest.param(FormatV04(), id="V04"), pytest.param(FormatV05(), id="V05")),
    )
    def test_finder(self, fmt):
        img_dir = (self.path / "images").mkdir()

        # test with empty directory - for code coverage
        finder(img_dir, 8000, True)
        assert not (img_dir / "biofile_finder.csv").exists()

        img_dir2 = (img_dir / "dir2").mkdir()
        bf2raw_dir = (img_dir / "bf2raw.zarr").mkdir()
        main(
            [
                "create",
                "--method=astronaut",
                (str(img_dir / "astronaut")),
                "--format",
                fmt.version,
            ]
        )
        main(
            [
                "create",
                "--method=coins",
                (str(img_dir2 / "coins")),
                "--format",
                fmt.version,
            ]
        )
        (bf2raw_dir / "OME").mkdir()

        # write minimal bioformats2raw and xml metadata
        with open(bf2raw_dir / ".zattrs", "w") as f:
            f.write("""{"bioformats2raw.layout" : 3}""")
        with open(bf2raw_dir / "OME" / "METADATA.ome.xml", "w") as f:
            f.write(
                """<?xml version="1.0" encoding="UTF-8"?>
                <OME><Image ID="Image:1" Name="test.fake"></Image></OME>
                """
            )

        # create a plate
        plate_path = Path(img_dir2.mkdir("plate"))
        store = parse_url(plate_path, mode="w", fmt=fmt).store
        root = zarr.group(store=store)
        write_plate_metadata(root, ["A"], ["1"], ["A/1"])

        finder(img_dir, 8000, True)

        assert (img_dir / "biofile_finder.csv").exists()
        csv_text = (img_dir / "biofile_finder.csv").read_text(encoding="utf-8")
        print(csv_text)
        assert "File Path,File Name,Folders,Uploaded" in csv_text
        assert "dir2/plate/A/1/0,plate,dir2" in csv_text
        assert "coins,dir2" in csv_text
        assert "test.fake" in csv_text

    def test_find_multiscales(self):
        # for code coverage...
        empty_dir = (self.path / "find_multiscales").mkdir()
        assert len(find_multiscales(empty_dir)) == 0



"""Reading logic for ome-zarr.

Primary entry point is the :func:`~ome_zarr.io.parse_url` method.
"""

import logging
from pathlib import Path
from urllib.parse import urljoin

import dask.array as da
import zarr
from zarr.storage import FsspecStore, LocalStore, StoreLike

from .format import CurrentFormat, Format, detect_format
from .types import JSONDict

LOGGER = logging.getLogger("ome_zarr.io")


class ZarrLocation:
    """
    IO primitive for reading and writing Zarr data. Uses a store for all
    data access.

    No assumptions about the existence of the given path string are made.
    Attempts are made to load various metadata files and cache them internally.
    """

    def __init__(
        self,
        path: StoreLike,
        mode: str = "r",
        fmt: Format = CurrentFormat(),
    ) -> None:
        LOGGER.debug("ZarrLocation.__init__ path: %s, fmt: %s", path, fmt.version)
        self.__fmt = fmt
        self.__mode = mode
        if isinstance(path, Path):
            self.__path = str(path.resolve())
        elif isinstance(path, str):
            self.__path = path
        elif isinstance(path, FsspecStore):
            self.__path = path.path
        elif isinstance(path, LocalStore):
            self.__path = str(path.root)
        else:
            raise TypeError(f"not expecting: {type(path)}")

        loader = fmt
        if loader is None:
            loader = CurrentFormat()
        self.__store: FsspecStore = (
            path
            if isinstance(path, (FsspecStore, LocalStore))
            else loader.init_store(self.__path, mode)
        )
        self.__init_metadata()
        detected = detect_format(self.__metadata, loader)
        LOGGER.debug("ZarrLocation.__init__ %s detected: %s", path, detected)
        if detected != fmt:
            LOGGER.warning(
                "version mismatch: detected: %s, requested: %s", detected, fmt
            )
            self.__fmt = detected
            self.__store = detected.init_store(self.__path, mode)
            self.__init_metadata()

    def __init_metadata(self) -> None:
        """
        Load the Zarr metadata files for the given location.
        """
        self.zgroup: JSONDict = {}
        self.zarray: JSONDict = {}
        self.__metadata: JSONDict = {}
        self.__exists: bool = True
        # If we want to *create* a new zarr v2 group, we need to specify
        # zarr_format. This is not needed for reading.
        zarr_format = None
        try:
            # this group is used to get zgroup metadata
            # used for info, download, Spec.match() via root_attrs() etc.
            # and to check if the group exists for reading. Only need "r" mode for this.
            group = zarr.open_group(
                store=self.__store, path="/", mode="r", zarr_format=zarr_format
            )
            self.zgroup = group.attrs.asdict()
            # For zarr v3, everything is under the "ome" namespace
            if "ome" in self.zgroup:
                self.zgroup = self.zgroup["ome"]
            self.__metadata = self.zgroup
        except (ValueError, FileNotFoundError):
            # group doesn't exist. If we are in "w" mode, we need to create it.
            if self.__mode == "w":
                # If we are creating a new group, we need to specify the zarr_format.
                zarr_format = self.__fmt.zarr_format
                group = zarr.open_group(
                    store=self.__store, path="/", mode="w", zarr_format=zarr_format
                )
            else:
                self.__exists = False

    def __repr__(self) -> str:
        """Print the path as well as whether this is a group or an array."""
        suffix = ""
        if self.zgroup:
            suffix += " [zgroup]"
        if self.zarray:
            suffix += " [zarray]"
        return f"{self.subpath('')}{suffix}"

    def exists(self) -> bool:
        """Return true if either zgroup or zarray metadata exists."""
        return self.__exists

    @property
    def fmt(self) -> Format:
        return self.__fmt

    @property
    def mode(self) -> str:
        return self.__mode

    @property
    def version(self) -> str:
        """Return the version of the OME-NGFF spec used for this location."""
        return self.__fmt.version

    @property
    def path(self) -> str:
        return self.__path

    @property
    def store(self) -> FsspecStore:
        """Return the initialized store for this location"""
        assert self.__store is not None
        return self.__store

    @property
    def root_attrs(self) -> JSONDict:
        """Return the contents of the zattrs file."""
        return dict(self.__metadata)

    def load(self, subpath: str = "") -> da.core.Array:
        """Use dask.array.from_zarr to load the subpath."""
        return da.from_zarr(self.__store, subpath)

    def __eq__(self, rhs: object) -> bool:
        if type(self) is not type(rhs):
            return False
        if not isinstance(rhs, ZarrLocation):
            return False
        return self.subpath() == rhs.subpath()

    def basename(self) -> str:
        """Return the last element of the underlying location.

        >>> ZarrLocation("/tmp/foo").basename()
        'foo'
        >>> ZarrLocation("https://example.com/bar").basename()
        'bar'
        >>> ZarrLocation("https://example.com/baz/").basename()
        'baz'
        """
        path = (self.__path.endswith("/") and self.__path[0:-1]) or self.__path
        return path.split("/")[-1]

    # TODO: update to from __future__ import annotations with 3.7+
    def create(self, path: str) -> "ZarrLocation":
        """Create a new Zarr location for the given path."""
        subpath = self.subpath(path)
        LOGGER.debug("open(%s(%s))", self.__class__.__name__, subpath)
        return self.__class__(subpath, mode=self.__mode, fmt=self.__fmt)

    def parts(self) -> list[str]:
        if self._isfile():
            return list(Path(self.__path).parts)
        else:
            return self.__path.split("/")

    def subpath(self, subpath: str = "") -> str:
        if self._isfile():
            filename = Path(self.__path) / subpath
            filename = filename.resolve()
            return str(filename)
        elif self._ishttp():
            url = str(self.__path)
            if not url.endswith("/"):
                url = f"{url}/"
            return urljoin(url, subpath)
        # Might require a warning
        elif self.__path.endswith("/"):
            return f"{self.__path}{subpath}"
        else:
            return f"{self.__path}/{subpath}"

    def _isfile(self) -> bool:
        """
        Return whether the current underlying implementation
        points to a local file or not.
        """
        return isinstance(self.__store, LocalStore)

    def _ishttp(self) -> bool:
        """
        Return whether the current underlying implementation
        points to a URL
        """
        return self.__store.fs.protocol in ["http", "https"]


def parse_url(
    path: Path | str, mode: str = "r", fmt: Format = CurrentFormat()
) -> ZarrLocation | None:
    """Convert a path string or URL to a ZarrLocation subclass.

    :param path: Path to parse.
    :param mode: Mode to open in.
    :param fmt: Version of the OME-NGFF spec to open path with.

    :return: `ZarrLocation`.
        If mode is 'r', and the path does not exist returns None.
        If there is an error opening the path, also returns None.

    >>> parse_url('does-not-exist')
    """
    loc = ZarrLocation(path, mode=mode, fmt=fmt)
    if "r" in mode and not loc.exists():
        return None
    else:
        return loc



import filecmp
import json
import pathlib
from typing import Any

import dask.array as da
import numpy as np
import pytest
import zarr
from dask import persist
from numcodecs import Blosc
from zarr.abc.codec import BytesBytesCodec
from zarr.codecs import BloscCodec

from ome_zarr.format import (
    CurrentFormat,
    FormatV01,
    FormatV02,
    FormatV03,
    FormatV04,
    FormatV05,
)
from ome_zarr.io import ZarrLocation, parse_url
from ome_zarr.reader import Multiscales, Reader
from ome_zarr.scale import Scaler
from ome_zarr.writer import (
    _get_valid_axes,
    _retuple,
    write_image,
    write_labels,
    write_multiscale,
    write_multiscale_labels,
    write_multiscales_metadata,
    write_plate_metadata,
    write_well_metadata,
)

TRANSFORMATIONS = [
    [{"scale": [1, 1, 0.5, 0.18, 0.18], "type": "scale"}],
    [{"scale": [1, 1, 0.5, 0.36, 0.36], "type": "scale"}],
    [{"scale": [1, 1, 0.5, 0.72, 0.72], "type": "scale"}],
    [{"scale": [1, 1, 0.5, 1.44, 1.44], "type": "scale"}],
    [{"scale": [1, 1, 0.5, 2.88, 2.88], "type": "scale"}],
]


class TestWriter:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))
        # create zarr v2 group...
        self.store = parse_url(self.path, mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)
        self.group = self.root.create_group("test")

        # let's create zarr v3 group too...
        self.path_v3 = self.path / "v3"
        store_v3 = parse_url(self.path_v3, mode="w").store
        root_v3 = zarr.group(store=store_v3)
        self.group_v3 = root_v3.create_group("test")

    def create_data(self, shape, dtype=np.uint8, mean_val=10):
        rng = np.random.default_rng(0)
        return rng.poisson(mean_val, size=shape).astype(dtype)

    @pytest.fixture(
        params=(
            (1, 2, 1, 256, 256),
            (3, 512, 512),
            (300, 500),  # test edge chunks of different shapes
        ),
        ids=["5D", "3D", "2D"],
    )
    def shape(self, request):
        return request.param

    @pytest.fixture(params=[True, False], ids=["scale", "noop"])
    def scaler(self, request):
        if request.param:
            return Scaler()
        else:
            return None

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV01, id="V01"),
            pytest.param(FormatV02, id="V02"),
            pytest.param(FormatV03, id="V03"),
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    @pytest.mark.parametrize("storage_options_list", [True, False])
    def test_writer(
        self, shape, scaler, format_version, array_constructor, storage_options_list
    ):
        version = format_version()

        if version.version == "0.5":
            group = self.group_v3
            grp_path = self.path_v3 / "test"
        else:
            group = self.group
            grp_path = self.path / "test"

        data = self.create_data(shape)
        data = array_constructor(data)
        axes = "tczyx"[-len(shape) :]
        transformations = []
        for dataset_transfs in TRANSFORMATIONS:
            transf = dataset_transfs[0]
            # e.g. slice [1, 1, z, x, y] -> [z, x, y] for 3D
            transformations.append(
                [{"type": "scale", "scale": transf["scale"][-len(shape) :]}]
            )
        if scaler is None:
            transformations = [transformations[0]]
        chunks = [(128, 128), (50, 50), (25, 25), (25, 25), (25, 25), (25, 25)]
        storage_options = {"chunks": chunks[0]}
        if storage_options_list:
            storage_options = [{"chunks": chunk} for chunk in chunks]
        write_image(
            image=data,
            group=group,
            scaler=scaler,
            fmt=version,
            axes=axes,
            coordinate_transformations=transformations,
            storage_options=storage_options,
        )

        # Verify
        reader = Reader(parse_url(f"{grp_path}"))
        node = next(iter(reader()))
        assert Multiscales.matches(node.zarr)
        if version.version in ("0.1", "0.2"):
            # v0.1 and v0.2 MUST be 5D
            assert node.data[0].ndim == 5
        else:
            assert node.data[0].shape == shape
        print("node.metadata", node.metadata)
        if version.version not in ("0.1", "0.2", "0.3"):
            for transf, expected in zip(
                node.metadata["coordinateTransformations"], transformations
            ):
                assert transf == expected
            assert len(node.metadata["coordinateTransformations"]) == len(node.data)
        # check chunks for first 2 resolutions (before shape gets smaller than chunk)
        for level, nd_array in enumerate(node.data[:2]):
            expected = chunks[level] if storage_options_list else chunks[0]
            first_chunk = [c[0] for c in nd_array.chunks]
            assert tuple(first_chunk) == _retuple(expected, nd_array.shape)
        assert np.allclose(data, node.data[0][...].compute())

    def test_mix_zarr_formats(self):
        # check group zarr v2 and v3 matches fmt
        data = self.create_data((64, 64, 64))
        with pytest.raises(ValueError, match=r"Group is zarr_format: 2"):
            write_image(data, self.group, axes="zyx", fmt=CurrentFormat())

        with pytest.raises(ValueError, match=r"Group is zarr_format: 3"):
            write_multiscale([data], self.group_v3, fmt=FormatV04())

        with pytest.raises(ValueError, match=r"Group is zarr_format: 3"):
            write_plate_metadata(self.group_v3, ["A"], ["1"], ["A/1"], fmt=FormatV04())

        with pytest.raises(ValueError, match=r"Group is zarr_format: 2"):
            write_well_metadata(self.group, [{"path": "0"}], fmt=CurrentFormat())

    @pytest.mark.parametrize("zarr_format", [2, 3])
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_write_image_current(self, array_constructor, zarr_format):
        shape = (64, 64, 64)
        data = self.create_data(shape)
        data = array_constructor(data)

        if zarr_format == 2:
            group = self.group
            grp_path = self.path / "test"
        else:
            group = self.group_v3
            grp_path = self.path_v3 / "test"

        write_image(data, group, axes="zyx")
        reader = Reader(parse_url(f"{grp_path}"))

        # manually check this is zarr v2 or v3
        if zarr_format == 2:
            json_text = (grp_path / ".zattrs").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text)
        else:
            json_text = (grp_path / "zarr.json").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text).get("attributes", {}).get("ome", {})
        assert "multiscales" in attrs_json

        image_node = next(iter(reader()))
        for transfs in image_node.metadata["coordinateTransformations"]:
            assert len(transfs) == 1
            assert transfs[0]["type"] == "scale"
            assert len(transfs[0]["scale"]) == len(shape)
            # Scaler only downsamples x and y. z scale will be 1
            assert transfs[0]["scale"][0] == 1
            for value in transfs[0]["scale"]:
                assert value >= 1

    @pytest.mark.parametrize("read_from_zarr", [True, False])
    @pytest.mark.parametrize("compute", [True, False])
    @pytest.mark.parametrize("zarr_format", [2, 3])
    def test_write_image_dask(self, read_from_zarr, compute, zarr_format):
        if zarr_format == 2:
            grp_path = self.path / "test"
            fmt = FormatV04()
            zarr_attrs = ".zattrs"
            zarr_array = ".zarray"
            group = self.group
        else:
            grp_path = self.path_v3 / "test"
            fmt = CurrentFormat()
            zarr_attrs = "zarr.json"
            zarr_array = "zarr.json"
            group = self.group_v3

        # Size 100 tests resize shapes: https://github.com/ome/ome-zarr-py/issues/219
        shape = (128, 200, 200)
        data = self.create_data(shape)
        data_delayed = da.from_array(data)
        chunks = (32, 32)
        # same NAME needed for exact zarr_attrs match below
        # (otherwise group.name is used)
        NAME = "test_write_image_dask"
        opts = {"chunks": chunks}
        if read_from_zarr:
            # write to zarr and re-read as dask...
            path = f"{grp_path}/temp/"
            store = parse_url(path, mode="w", fmt=fmt).store
            # store and group will be zarr v2 or v3 depending on fmt
            temp_group = zarr.group(store=store).create_group("to_dask")
            assert temp_group.info._zarr_format == zarr_format
            write_image(
                data_delayed,
                temp_group,
                axes="zyx",
                storage_options=opts,
                name=NAME,
            )
            print("PATH", f"{grp_path}/temp/to_dask")
            loc = ZarrLocation(f"{grp_path}/temp/to_dask")

            reader = Reader(loc)()
            nodes = list(reader)
            data_delayed = nodes[0].load(Multiscales).array(resolution="0")
            # check that the data is the same
            assert np.allclose(data, data_delayed[...].compute())

        assert group.info._zarr_format == zarr_format
        dask_delayed_jobs = write_image(
            data_delayed,
            group,
            axes="zyx",
            storage_options={"chunks": chunks},
            compute=compute,
            name=NAME,
        )

        assert not compute == len(dask_delayed_jobs)

        if not compute:
            # can be configured to use a Local or Slurm cluster
            # before persisting the jobs
            dask_delayed_jobs = persist(*dask_delayed_jobs)

        # check the data written to zarr v2 or v3 group
        reader = Reader(parse_url(f"{grp_path}"))
        image_node = next(iter(reader()))
        first_chunk = [c[0] for c in image_node.data[0].chunks]
        assert tuple(first_chunk) == _retuple(chunks, image_node.data[0].shape)
        for level, transfs in enumerate(
            image_node.metadata["coordinateTransformations"]
        ):
            assert len(transfs) == 1
            assert transfs[0]["type"] == "scale"
            assert len(transfs[0]["scale"]) == len(shape)
            # Scaler only downsamples x and y. z scale will be 1
            assert transfs[0]["scale"][0] == 1
            for value in transfs[0]["scale"]:
                assert value >= 1
            if read_from_zarr and level < 3:
                # if shape smaller than chunk, dask writer uses chunk == shape
                # so we only compare larger resolutions
                assert filecmp.cmp(
                    f"{grp_path}/temp/to_dask/{level}/{zarr_array}",
                    f"{grp_path}/{level}/{zarr_array}",
                    shallow=False,
                )

        if read_from_zarr:
            # exact match, including NAME
            assert filecmp.cmp(
                f"{grp_path}/temp/to_dask/{zarr_attrs}",
                f"{grp_path}/{zarr_attrs}",
                shallow=False,
            )

    def test_write_image_scalar_chunks(self):
        """
        Make sure a scalar chunks value is applied to all dimensions,
        matching the behaviour of zarr-python.
        """
        shape = (64, 64, 64)
        data = np.array(self.create_data(shape))
        write_image(
            image=data, group=self.group, axes="xyz", storage_options={"chunks": 32}
        )
        for data in self.group.array_values():
            print(data)
            assert data.chunks == (32, 32, 32)

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_write_image_compressed(self, array_constructor, format_version):
        shape = (64, 64, 64)
        data = self.create_data(shape)
        data = array_constructor(data)
        path = self.path / "test_write_image_compressed"
        store = parse_url(path, mode="w", fmt=format_version()).store
        root = zarr.group(store=store)
        CNAME = "lz4"
        LEVEL = 4
        if format_version().zarr_format == 3:
            compressor = BloscCodec(cname=CNAME, clevel=LEVEL, shuffle="shuffle")
            assert isinstance(compressor, BytesBytesCodec)
            if isinstance(data, da.Array):
                # skip test - can't get this to pass. Fails with:
                # ValueError: compressor cannot be used for arrays with zarr_format 3.
                # Use bytes-to-bytes codecs instead.
                pytest.skip("storage_options['compressor'] fails in da.to_zarr()")
        else:
            compressor = Blosc(cname=CNAME, clevel=LEVEL, shuffle=Blosc.SHUFFLE)

        write_image(
            data,
            root,
            axes="zyx",
            storage_options={"compressor": compressor},
        )
        group = zarr.open(f"{path}")
        for ds in ["0", "1"]:
            assert len(group[ds].info._compressors) > 0
            comp = group[ds].info._compressors[0]
            if format_version().zarr_format == 3:
                print("comp", comp.to_dict())
                # {'configuration': {'checksum': False, 'level': 0}, 'name': 'zstd'}
                assert comp.to_dict() == {
                    "name": "blosc",
                    "configuration": {
                        "typesize": 1,
                        "cname": CNAME,
                        "clevel": LEVEL,
                        "shuffle": "shuffle",
                        "blocksize": 0,
                    },
                }
            else:
                print("comp", comp.get_config())
                assert comp.get_config() == {
                    "id": "blosc",
                    "cname": CNAME,
                    "clevel": LEVEL,
                    "shuffle": Blosc.SHUFFLE,
                    "blocksize": 0,
                }

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_default_compression(self, array_constructor, format_version):
        """Test that the default compression is not None.

        We make an array of zeros which should compress trivially easily,
        write out the chunks, and check that they are smaller than the raw
        data.
        """
        arr_np = np.zeros((2, 50, 200, 400), dtype=np.uint8)
        # avoid empty chunks so they are guaranteed to be written out to disk
        arr_np[0, 0, 0, 0] = 1
        # 4MB chunks, trivially compressible
        arr = array_constructor(arr_np)
        # tempdir = TemporaryDirectory(suffix=".ome.zarr")
        # self.path = pathlib.Path(tmpdir.mkdir("data"))
        path = self.path / "test_default_compression"
        store = parse_url(path, mode="w", fmt=format_version()).store
        root = zarr.group(store=store)
        assert root.info._zarr_format == format_version().zarr_format
        # no compressor options, we are checking default
        write_image(
            arr, group=root, axes="tzyx", storage_options=dict(chunks=(1, 100, 100))
        )

        # check chunk: multiscale level 0, 4D chunk at (0, 0, 0, 0)
        c = ""
        for ds in ["0", "1"]:
            if format_version().zarr_format == 3:
                assert (path / "zarr.json").exists()
                assert (path / ds / "zarr.json").exists()
                c = "c/"
                json_text = (path / ds / "zarr.json").read_text(encoding="utf-8")
                arr_json = json.loads(json_text)
                assert arr_json["codecs"][0]["name"] == "bytes"
                assert arr_json["codecs"][1] == {
                    "name": "zstd",
                    "configuration": {"level": 0, "checksum": False},
                }
            else:
                assert (path / ".zattrs").exists()
                json_text = (path / ds / ".zarray").read_text(encoding="utf-8")
                arr_json = json.loads(json_text)
                assert arr_json["compressor"] == {
                    "blocksize": 0,
                    "clevel": 5,
                    "cname": "zstd",
                    "id": "blosc",
                    "shuffle": 1,
                }

        chunk_size = (path / f"0/{c}0/0/0/0").stat().st_size
        assert chunk_size < 4e6

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    def test_validate_coordinate_transforms(self, format_version):
        fmt = format_version()

        transformations = [
            [{"type": "scale", "scale": (1, 1)}],
            [{"type": "scale", "scale": (0.5, 0.5)}],
        ]
        fmt.validate_coordinate_transformations(2, 2, transformations)

        with pytest.raises(ValueError):
            # transformations different length than levels
            fmt.validate_coordinate_transformations(2, 1, transformations)

        with pytest.raises(ValueError):
            transf = [[{"type": "scale", "scale": ("1", 1)}]]
            fmt.validate_coordinate_transformations(2, 1, transf)

        with pytest.raises(ValueError):
            transf = [[{"type": "foo", "scale": (1, 1)}]]
            fmt.validate_coordinate_transformations(2, 1, transf)

        with pytest.raises(ValueError):
            # scale list of floats different length from 3
            transf = [[{"type": "scale", "scale": (1, 1)}]]
            fmt.validate_coordinate_transformations(3, 1, transf)

        translate = [{"type": "translation", "translation": (1, 1)}]
        scale_then_trans = [transf + translate for transf in transformations]

        fmt.validate_coordinate_transformations(2, 2, scale_then_trans)

        trans_then_scale = [translate + transf for transf in transformations]
        with pytest.raises(ValueError):
            # scale must come first
            fmt.validate_coordinate_transformations(2, 2, trans_then_scale)

        with pytest.raises(ValueError):
            scale_then_trans2 = [transf + translate for transf in scale_then_trans]
            # more than 1 transformation
            fmt.validate_coordinate_transformations(2, 2, scale_then_trans2)

    def test_dim_names(self):
        v03 = FormatV03()

        # v0.3 MUST specify axes for 3D or 4D data
        with pytest.raises(ValueError):
            _get_valid_axes(3, axes=None, fmt=v03)

        # ndims must match axes length
        with pytest.raises(ValueError):
            _get_valid_axes(3, axes="yx", fmt=v03)

        # axes must be ordered tczyx
        with pytest.raises(ValueError):
            _get_valid_axes(3, axes="yxt", fmt=v03)
        with pytest.raises(ValueError):
            _get_valid_axes(2, axes=["x", "y"], fmt=v03)
        with pytest.raises(ValueError):
            _get_valid_axes(5, axes="xyzct", fmt=v03)

        # valid axes - no change, converted to list
        assert _get_valid_axes(2, axes=["y", "x"], fmt=v03) == ["y", "x"]
        assert _get_valid_axes(5, axes="tczyx", fmt=v03) == [
            "t",
            "c",
            "z",
            "y",
            "x",
        ]

        # if 2D or 5D, axes can be assigned automatically
        assert _get_valid_axes(2, axes=None, fmt=v03) == ["y", "x"]
        assert _get_valid_axes(5, axes=None, fmt=v03) == ["t", "c", "z", "y", "x"]

        # for v0.1 or v0.2, axes should be None
        assert _get_valid_axes(2, axes=["y", "x"], fmt=FormatV01()) is None
        assert _get_valid_axes(2, axes=["y", "x"], fmt=FormatV02()) is None

        # check that write_image is checking axes
        data = self.create_data((125, 125))
        with pytest.raises(ValueError):
            write_image(
                image=data,
                group=self.group,
                fmt=v03,
                axes="xyz",
            )

    def test_axes_dicts(self):
        v04 = FormatV04()

        # ALL axes must specify 'name'
        with pytest.raises(ValueError):
            _get_valid_axes(2, axes=[{"name": "y"}, {}], fmt=v04)

        all_dims = [
            {"name": "t", "type": "time"},
            {"name": "c", "type": "channel"},
            {"name": "z", "type": "space"},
            {"name": "y", "type": "space"},
            {"name": "x", "type": "space"},
        ]

        # auto axes for 2D, 5D, converted to dict for v0.4
        assert _get_valid_axes(2, axes=None, fmt=v04) == all_dims[-2:]
        assert _get_valid_axes(5, axes=None, fmt=v04) == all_dims

        # convert from list or string
        assert _get_valid_axes(3, axes=["z", "y", "x"], fmt=v04) == all_dims[-3:]
        assert _get_valid_axes(4, axes="czyx", fmt=v04) == all_dims[-4:]

        # invalid based on ordering of types
        with pytest.raises(ValueError):
            assert _get_valid_axes(3, axes=["y", "c", "x"], fmt=v04)
        with pytest.raises(ValueError):
            assert _get_valid_axes(4, axes="ctyx", fmt=v04)

        # custom types
        assert _get_valid_axes(3, axes=["foo", "y", "x"], fmt=v04) == [
            {"name": "foo"},
            all_dims[-2],
            all_dims[-1],
        ]

        # space types can be in ANY order
        assert _get_valid_axes(3, axes=["x", "z", "y"], fmt=v04) == [
            all_dims[-1],
            all_dims[-3],
            all_dims[-2],
        ]

        # Not allowed multiple custom types
        with pytest.raises(ValueError):
            _get_valid_axes(4, axes=["foo", "bar", "y", "x"], fmt=v04)

        # unconventional naming is allowed
        strange_axes = [
            {"name": "duration", "type": "time"},
            {"name": "rotation", "type": "angle"},
            {"name": "dz", "type": "space"},
            {"name": "WIDTH", "type": "space"},
        ]
        assert _get_valid_axes(4, axes=strange_axes, fmt=v04) == strange_axes

        # check that write_image is checking axes
        data = self.create_data((125, 125))
        with pytest.raises(ValueError):
            write_image(
                image=data,
                group=self.group,
                fmt=v04,
                axes="xt",
            )


class TestMultiscalesMetadata:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))
        # create zarr v2 group...
        self.store = parse_url(self.path, mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)

        # let's create zarr v3 group too...
        self.path_v3 = self.path / "v3"
        store_v3 = parse_url(self.path_v3, mode="w").store
        self.root_v3 = zarr.group(store=store_v3)

    @pytest.mark.parametrize("fmt", (FormatV04(), FormatV05()))
    def test_multi_levels_transformations(self, fmt):
        datasets = []
        for level, transf in enumerate(TRANSFORMATIONS):
            datasets.append({"path": str(level), "coordinateTransformations": transf})
        if fmt.version == "0.5":
            group = self.root_v3
        else:
            group = self.root
        write_multiscales_metadata(group, datasets, axes="tczyx")
        # we want to be sure this is zarr v2 / v3
        attrs = group.attrs
        if fmt.version == "0.5":
            attrs = attrs.get("ome")
            assert "version" in attrs
            json_text = (self.path_v3 / "zarr.json").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text).get("attributes", {}).get("ome", {})
        else:
            json_text = (self.path / ".zattrs").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text)
            assert "version" in attrs["multiscales"][0]
        assert "multiscales" in attrs_json
        assert "multiscales" in attrs
        assert attrs["multiscales"][0]["datasets"] == datasets

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02(), FormatV03()))
    def test_version(self, fmt):
        write_multiscales_metadata(self.root, [{"path": "0"}], fmt=fmt)
        assert "multiscales" in self.root.attrs
        assert self.root.attrs["multiscales"][0]["version"] == fmt.version
        assert self.root.attrs["multiscales"][0]["datasets"] == [{"path": "0"}]

    @pytest.mark.parametrize(
        "axes",
        (
            ["y", "x"],
            ["c", "y", "x"],
            ["z", "y", "x"],
            ["t", "y", "x"],
            ["t", "c", "y", "x"],
            ["t", "z", "y", "x"],
            ["c", "z", "y", "x"],
            ["t", "c", "z", "y", "x"],
        ),
    )
    def test_axes_V03(self, axes):
        write_multiscales_metadata(
            self.root, [{"path": "0"}], fmt=FormatV03(), axes=axes
        )
        assert "multiscales" in self.root.attrs
        # for v0.3, axes is a list of names
        assert self.root.attrs["multiscales"][0]["axes"] == axes
        with pytest.raises(ValueError):
            # for v0.4 and above, paths no-longer supported (need dataset dicts)
            write_multiscales_metadata(self.root, ["0"], axes=axes, fmt=FormatV04())

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02()))
    def test_axes_ignored(self, fmt):
        write_multiscales_metadata(
            self.root, [{"path": "0"}], fmt=fmt, axes=["t", "c", "z", "y", "x"]
        )
        assert "multiscales" in self.root.attrs
        assert "axes" not in self.root.attrs["multiscales"][0]

    @pytest.mark.parametrize(
        "axes",
        (
            [],
            ["i", "j"],
            ["x", "y"],
            ["y", "x", "c"],
            ["x", "y", "z", "c", "t"],
        ),
    )
    def test_invalid_0_3_axes(self, axes):
        with pytest.raises(ValueError):
            write_multiscales_metadata(self.root, ["0"], fmt=FormatV03(), axes=axes)

    @pytest.mark.parametrize("datasets", ([], None, "0", ["0"], [{"key": 1}]))
    def test_invalid_datasets(self, datasets):
        with pytest.raises(ValueError):
            write_multiscales_metadata(
                self.root, datasets, axes=["t", "c", "z", "y", "x"], fmt=FormatV04()
            )

    @pytest.mark.parametrize(
        "coordinateTransformations",
        (
            [{"type": "scale", "scale": [1, 1]}],
            [
                {"type": "scale", "scale": [1, 1]},
                {"type": "translation", "translation": [0, 0]},
            ],
        ),
    )
    def test_valid_transformations(self, coordinateTransformations):
        axes = [{"name": "y", "type": "space"}, {"name": "x", "type": "space"}]
        datasets = [
            {
                "path": "0",
                "coordinateTransformations": coordinateTransformations,
            }
        ]
        write_multiscales_metadata(self.root, datasets, axes=axes, fmt=FormatV04())
        assert "multiscales" in self.root.attrs
        assert self.root.attrs["multiscales"][0]["axes"] == axes
        assert self.root.attrs["multiscales"][0]["datasets"] == datasets

    @pytest.mark.parametrize(
        "coordinateTransformations",
        (
            [],
            None,
            [{"type": "scale"}],
            [{"scale": [1, 1]}],
            [{"type": "scale", "scale": ["1", 1]}],
            [{"type": "scale", "scale": [1, 1, 1]}],
            [{"type": "scale", "scale": [1, 1]}, {"type": "scale", "scale": [1, 1]}],
            [
                {"type": "scale", "scale": [1, 1]},
                {"type": "translation", "translation": ["0", 0]},
            ],
            [
                {"type": "translation", "translation": [0, 0]},
            ],
            [
                {"type": "scale", "scale": [1, 1]},
                {"type": "translation", "translation": [0, 0, 0]},
            ],
            [
                {"type": "translation", "translation": [0, 0]},
                {"type": "scale", "scale": [1, 1]},
            ],
            [
                {"type": "scale", "scale": [1, 1]},
                {"type": "translation", "translation": [0, 0]},
                {"type": "translation", "translation": [1, 0]},
            ],
            [
                {"type": "scale", "scale": [1, 1]},
                {"translation": [0, 0]},
            ],
            [
                {"type": "scale", "scale": [1, 1]},
                {"type": "translation", "translate": [0, 0]},
            ],
        ),
    )
    def test_invalid_transformations(self, coordinateTransformations):
        axes = [{"name": "y", "type": "space"}, {"name": "x", "type": "space"}]
        datasets = [
            {"path": "0", "coordinateTransformations": coordinateTransformations}
        ]
        with pytest.raises(ValueError):
            write_multiscales_metadata(self.root, datasets, axes=axes, fmt=FormatV04())

    @pytest.mark.parametrize(
        "metadata",
        [
            {
                "channels": [
                    {
                        "color": "FF0000",
                        "window": {"start": 0, "end": 255, "min": 0, "max": 255},
                    }
                ]
            },
            {"channels": [{"color": "FF0000"}]},
            {"channels": [{"color": "FF000"}]},  # test wrong metadata
            {"channels": [{"window": []}]},  # test wrong metadata
            {
                "channels": [  # test wrong metadata
                    {"color": "FF0000", "window": {"start": 0, "end": 255, "min": 0}},
                ]
            },
            None,
        ],
    )
    def test_omero_metadata(self, metadata: dict[str, Any] | None):
        datasets = []
        for level, transf in enumerate(TRANSFORMATIONS):
            datasets.append({"path": str(level), "coordinateTransformations": transf})
        if metadata is None:
            with pytest.raises(
                KeyError, match="If `'omero'` is present, value cannot be `None`."
            ):
                write_multiscales_metadata(
                    self.root,
                    datasets,
                    axes="tczyx",
                    metadata={"omero": metadata},
                )
        else:
            window_metadata = (
                metadata["channels"][0].get("window")
                if "window" in metadata["channels"][0]
                else None
            )
            color_metadata = (
                metadata["channels"][0].get("color")
                if "color" in metadata["channels"][0]
                else None
            )
            if window_metadata is not None and len(window_metadata) < 4:
                if isinstance(window_metadata, dict):
                    with pytest.raises(KeyError, match=".*`'window'`.*"):
                        write_multiscales_metadata(
                            self.root,
                            datasets,
                            axes="tczyx",
                            metadata={"omero": metadata},
                            fmt=FormatV04(),
                        )
                elif isinstance(window_metadata, list):
                    with pytest.raises(TypeError, match=".*`'window'`.*"):
                        write_multiscales_metadata(
                            self.root,
                            datasets,
                            axes="tczyx",
                            metadata={"omero": metadata},
                            fmt=FormatV04(),
                        )
            elif color_metadata is not None and len(color_metadata) != 6:
                with pytest.raises(TypeError, match=".*`'color'`.*"):
                    write_multiscales_metadata(
                        self.root,
                        datasets,
                        axes="tczyx",
                        metadata={"omero": metadata},
                    )
            else:
                write_multiscales_metadata(
                    self.root,
                    datasets,
                    axes="tczyx",
                    metadata={"omero": metadata},
                )


class TestPlateMetadata:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))
        # create zarr v2 group...
        self.store = parse_url(self.path, mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)
        # create zarr v3 group...
        self.path_v3 = self.path / "v3"
        store_v3 = parse_url(self.path_v3, mode="w").store
        self.root_v3 = zarr.group(store=store_v3)

    @pytest.mark.parametrize("fmt", (FormatV04(), FormatV05()))
    def test_minimal_plate(self, fmt):
        if fmt.version == "0.4":
            group = self.root
        else:
            group = self.root_v3
        write_plate_metadata(group, ["A"], ["1"], ["A/1"])
        attrs = group.attrs
        if fmt.version != "0.4":
            attrs = attrs["ome"]
            assert attrs["version"] == fmt.version
        else:
            assert attrs["plate"]["version"] == fmt.version

        assert "plate" in attrs
        assert attrs["plate"]["columns"] == [{"name": "1"}]
        assert attrs["plate"]["rows"] == [{"name": "A"}]
        assert attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0}
        ]
        assert "name" not in attrs["plate"]
        assert "field_count" not in attrs["plate"]
        assert "acquisitions" not in attrs["plate"]

    @pytest.mark.parametrize("fmt", (FormatV04(), FormatV05()))
    def test_12wells_plate(self, fmt):
        rows = ["A", "B", "C", "D"]
        cols = ["1", "2", "3"]
        wells = [
            "A/1",
            "A/2",
            "A/3",
            "B/1",
            "B/2",
            "B/3",
            "C/1",
            "C/2",
            "C/3",
            "D/1",
            "D/2",
            "D/3",
        ]
        if fmt.version == "0.4":
            group = self.root
        else:
            group = self.root_v3
        write_plate_metadata(group, rows, cols, wells)
        attrs = group.attrs
        if fmt.version != "0.4":
            attrs = attrs["ome"]

        assert "plate" in attrs
        assert attrs["plate"]["columns"] == [
            {"name": "1"},
            {"name": "2"},
            {"name": "3"},
        ]
        assert attrs["plate"]["rows"] == [
            {"name": "A"},
            {"name": "B"},
            {"name": "C"},
            {"name": "D"},
        ]
        assert attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0},
            {"path": "A/2", "rowIndex": 0, "columnIndex": 1},
            {"path": "A/3", "rowIndex": 0, "columnIndex": 2},
            {"path": "B/1", "rowIndex": 1, "columnIndex": 0},
            {"path": "B/2", "rowIndex": 1, "columnIndex": 1},
            {"path": "B/3", "rowIndex": 1, "columnIndex": 2},
            {"path": "C/1", "rowIndex": 2, "columnIndex": 0},
            {"path": "C/2", "rowIndex": 2, "columnIndex": 1},
            {"path": "C/3", "rowIndex": 2, "columnIndex": 2},
            {"path": "D/1", "rowIndex": 3, "columnIndex": 0},
            {"path": "D/2", "rowIndex": 3, "columnIndex": 1},
            {"path": "D/3", "rowIndex": 3, "columnIndex": 2},
        ]
        assert "name" not in attrs["plate"]
        assert "field_count" not in attrs["plate"]
        assert "acquisitions" not in attrs["plate"]

    @pytest.mark.parametrize("fmt", (FormatV04(), FormatV05()))
    def test_sparse_plate(self, fmt):
        rows = ["A", "B", "C", "D", "E"]
        cols = ["1", "2", "3", "4", "5"]
        wells = [
            "B/2",
            "E/5",
        ]
        if fmt.version == "0.4":
            group = self.root
        else:
            group = self.root_v3
        write_plate_metadata(group, rows, cols, wells)
        attrs = group.attrs
        if fmt.version != "0.4":
            attrs = attrs["ome"]
        assert "plate" in attrs
        assert attrs["plate"]["columns"] == [
            {"name": "1"},
            {"name": "2"},
            {"name": "3"},
            {"name": "4"},
            {"name": "5"},
        ]
        assert attrs["plate"]["rows"] == [
            {"name": "A"},
            {"name": "B"},
            {"name": "C"},
            {"name": "D"},
            {"name": "E"},
        ]
        assert attrs["plate"]["wells"] == [
            {"path": "B/2", "rowIndex": 1, "columnIndex": 1},
            {"path": "E/5", "rowIndex": 4, "columnIndex": 4},
        ]
        assert "name" not in attrs["plate"]
        assert "field_count" not in attrs["plate"]
        assert "acquisitions" not in attrs["plate"]

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02(), FormatV03()))
    def test_legacy_wells(self, fmt):
        write_plate_metadata(self.root, ["A"], ["1"], ["A/1"], fmt=fmt)
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}]
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}]
        assert self.root.attrs["plate"]["version"] == fmt.version
        assert self.root.attrs["plate"]["wells"] == [{"path": "A/1"}]
        assert "name" not in self.root.attrs["plate"]
        assert "field_count" not in self.root.attrs["plate"]
        assert "acquisitions" not in self.root.attrs["plate"]

    def test_plate_name(self):
        # We don't need to test v04 and v05 for all tests since
        # the metadata is the same
        write_plate_metadata(self.root_v3, ["A"], ["1"], ["A/1"], name="test")
        attrs = self.root_v3.attrs["ome"]
        assert "plate" in attrs
        assert attrs["plate"]["columns"] == [{"name": "1"}]
        assert attrs["plate"]["name"] == "test"
        assert attrs["plate"]["rows"] == [{"name": "A"}]
        assert attrs["version"] == FormatV05().version
        assert attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0}
        ]
        assert "field_count" not in attrs["plate"]
        assert "acquisitions" not in attrs["plate"]

    def test_field_count(self):
        write_plate_metadata(
            self.root, ["A"], ["1"], ["A/1"], field_count=10, fmt=FormatV04()
        )
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}]
        assert self.root.attrs["plate"]["field_count"] == 10
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}]
        assert self.root.attrs["plate"]["version"] == FormatV04().version
        assert self.root.attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0}
        ]
        assert "name" not in self.root.attrs["plate"]
        assert "acquisitions" not in self.root.attrs["plate"]

    def test_acquisitions_minimal(self):
        a = [{"id": 1}, {"id": 2}, {"id": 3}]
        write_plate_metadata(
            self.root, ["A"], ["1"], ["A/1"], acquisitions=a, fmt=FormatV04()
        )
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["acquisitions"] == a
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}]
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}]
        assert self.root.attrs["plate"]["version"] == FormatV04().version
        assert self.root.attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0}
        ]
        assert "name" not in self.root.attrs["plate"]
        assert "field_count" not in self.root.attrs["plate"]

    def test_acquisitions_maximal(self):
        a = [
            {
                "id": 1,
                "name": "acquisition_1",
                "description": " first acquisition",
                "maximumfieldcount": 2,
                "starttime": 1343749391000,
                "endtime": 1343749392000,
            }
        ]
        write_plate_metadata(
            self.root, ["A"], ["1"], ["A/1"], acquisitions=a, fmt=FormatV04()
        )
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["acquisitions"] == a
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}]
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}]
        assert self.root.attrs["plate"]["version"] == FormatV04().version
        assert self.root.attrs["plate"]["wells"] == [
            {"path": "A/1", "rowIndex": 0, "columnIndex": 0}
        ]
        assert "name" not in self.root.attrs["plate"]
        assert "field_count" not in self.root.attrs["plate"]

    @pytest.mark.parametrize(
        "acquisitions",
        (
            [0, 1],
            [{"name": "0"}, {"name": "1"}],
            [{"id": "0"}, {"id": "1"}],
        ),
    )
    def test_invalid_acquisition_keys(self, acquisitions):
        with pytest.raises(ValueError):
            write_plate_metadata(
                self.root_v3, ["A"], ["1"], ["A/1"], acquisitions=acquisitions
            )

    def test_unspecified_acquisition_keys(self):
        a = [{"id": 0, "unspecified_key": "0"}]
        write_plate_metadata(self.root, ["A"], ["1"], ["A/1"], acquisitions=a)
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["acquisitions"] == a

    @pytest.mark.parametrize(
        "wells",
        (None, [], [1]),
    )
    def test_invalid_well_list(self, wells):
        with pytest.raises(ValueError):
            write_plate_metadata(self.root, ["A"], ["1"], wells)

    @pytest.mark.parametrize(
        "wells",
        (
            # Missing required keys
            [{"id": "test"}],
            [{"path": "A/1"}],
            [{"path": "A/1", "rowIndex": 0}],
            [{"path": "A/1", "columnIndex": 0}],
            [{"rowIndex": 0, "columnIndex": 0}],
            # Invalid paths
            [{"path": 0, "rowIndex": 0, "columnIndex": 0}],
            [{"path": None, "rowIndex": 0, "columnIndex": 0}],
            [{"path": "plate/A/1", "rowIndex": 0, "columnIndex": 0}],
            [{"path": "plate/A1", "rowIndex": 0, "columnIndex": 0}],
            [{"path": "A/1/0", "rowIndex": 0, "columnIndex": 0}],
            [{"path": "A1", "rowIndex": 0, "columnIndex": 0}],
            [{"path": "0", "rowIndex": 0, "columnIndex": 0}],
            # Invalid row/column indices
            [{"path": "A/1", "rowIndex": "0", "columnIndex": 0}],
            [{"path": "A/1", "rowIndex": 0, "columnIndex": "0"}],
            # Undefined rows/columns
            [{"path": "C/1", "rowIndex": 2, "columnIndex": 0}],
            [{"path": "A/3", "rowIndex": 0, "columnIndex": 2}],
            # Mismatching indices
            [{"path": "A/1", "rowIndex": 0, "columnIndex": 1}],
            [{"path": "A/1", "rowIndex": 1, "columnIndex": 0}],
        ),
    )
    def test_invalid_well_keys(self, wells):
        with pytest.raises(ValueError):
            write_plate_metadata(self.root, ["A"], ["1"], wells, fmt=FormatV04())

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02(), FormatV03()))
    def test_legacy_unspecified_well_keys(self, fmt):
        wells = [
            {"path": "A/1", "unspecified_key": "alpha"},
            {"path": "A/2", "unspecified_key": "beta"},
            {"path": "B/1", "unspecified_key": "gamma"},
        ]
        write_plate_metadata(self.root, ["A", "B"], ["1", "2"], wells, fmt=fmt)
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}, {"name": "2"}]
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}, {"name": "B"}]
        assert self.root.attrs["plate"]["version"] == fmt.version
        assert self.root.attrs["plate"]["wells"] == wells

    def test_unspecified_well_keys(self):
        wells = [
            {
                "path": "A/1",
                "rowIndex": 0,
                "columnIndex": 0,
                "unspecified_key": "alpha",
            },
            {"path": "A/2", "rowIndex": 0, "columnIndex": 1, "unspecified_key": "beta"},
            {
                "path": "B/1",
                "rowIndex": 1,
                "columnIndex": 0,
                "unspecified_key": "gamma",
            },
        ]
        write_plate_metadata(self.root, ["A", "B"], ["1", "2"], wells, fmt=FormatV04())
        assert "plate" in self.root.attrs
        assert self.root.attrs["plate"]["columns"] == [{"name": "1"}, {"name": "2"}]
        assert self.root.attrs["plate"]["rows"] == [{"name": "A"}, {"name": "B"}]
        assert self.root.attrs["plate"]["version"] == FormatV04().version
        assert self.root.attrs["plate"]["wells"] == wells

    def test_missing_well_keys(self):
        wells = [
            {"path": "A/1"},
            {"path": "A/2"},
            {"path": "B/1"},
        ]
        with pytest.raises(ValueError):
            write_plate_metadata(
                self.root, ["A", "B"], ["1", "2"], wells, fmt=FormatV04()
            )

    def test_well_not_in_rows(self):
        wells = ["A/1", "B/1", "C/1"]
        with pytest.raises(ValueError):
            write_plate_metadata(
                self.root, ["A", "B"], ["1", "2"], wells, fmt=FormatV04()
            )

    def test_well_not_in_columns(self):
        wells = ["A/1", "A/2", "A/3"]
        with pytest.raises(ValueError):
            write_plate_metadata(
                self.root, ["A", "B"], ["1", "2"], wells, fmt=FormatV04()
            )

    @pytest.mark.parametrize("rows", (["A", "B", "B"], ["A", "&"]))
    def test_invalid_rows(self, rows):
        with pytest.raises(ValueError):
            write_plate_metadata(self.root, rows, ["1"], ["A/1"], fmt=FormatV04())

    @pytest.mark.parametrize("columns", (["1", "2", "2"], ["1", "&"]))
    def test_invalid_columns(self, columns):
        with pytest.raises(ValueError):
            write_plate_metadata(self.root, ["A"], columns, ["A/1"], fmt=FormatV04())


class TestWellMetadata:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))
        # create zarr v2 group...
        self.store = parse_url(self.path, mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)

        # create zarr v3 group too...
        self.path_v3 = self.path / "v3"
        store_v3 = parse_url(self.path_v3, mode="w").store
        self.root_v3 = zarr.group(store=store_v3)

    @pytest.mark.parametrize("fmt", (FormatV04(), FormatV05()))
    @pytest.mark.parametrize("images", (["0"], [{"path": "0"}]))
    def test_minimal_well(self, images, fmt):
        if fmt.version == "0.5":
            group = self.root_v3
        else:
            group = self.root
        write_well_metadata(group, images)
        # we want to be sure this is zarr v2 / v3, so we load json manually too
        attrs = group.attrs
        if fmt.version == "0.5":
            attrs = attrs.get("ome")
            assert attrs["version"] == fmt.version
            json_text = (self.path_v3 / "zarr.json").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text).get("attributes", {}).get("ome", {})
        else:
            json_text = (self.path / ".zattrs").read_text(encoding="utf-8")
            attrs_json = json.loads(json_text)
            assert attrs["well"]["version"] == fmt.version

        assert "well" in attrs_json
        assert attrs["well"]["images"] == [{"path": "0"}]

    @pytest.mark.parametrize(
        "images",
        (
            ["0", "1", "2"],
            [
                {"path": "0"},
                {"path": "1"},
                {"path": "2"},
            ],
        ),
    )
    def test_multiple_images(self, images):
        write_well_metadata(self.root_v3, images)
        assert "well" in self.root_v3.attrs.get("ome", {})
        assert self.root_v3.attrs["ome"]["well"]["images"] == [
            {"path": "0"},
            {"path": "1"},
            {"path": "2"},
        ]
        assert self.root_v3.attrs["ome"]["version"] == FormatV05().version

    @pytest.mark.parametrize("fmt", (FormatV01(), FormatV02(), FormatV03()))
    def test_version(self, fmt):
        write_well_metadata(self.root, ["0"], fmt=fmt)
        assert "well" in self.root.attrs
        assert self.root.attrs["well"]["images"] == [{"path": "0"}]
        assert self.root.attrs["well"]["version"] == fmt.version

    def test_multiple_acquisitions(self):
        images = [
            {"path": "0", "acquisition": 1},
            {"path": "1", "acquisition": 2},
            {"path": "2", "acquisition": 3},
        ]
        write_well_metadata(self.root, images)
        assert "well" in self.root.attrs
        assert self.root.attrs["well"]["images"] == images
        assert self.root.attrs["well"]["version"] == FormatV04().version

    @pytest.mark.parametrize(
        "images",
        (
            [{"acquisition": 0}, {"acquisition": 1}],
            [{"path": "0", "acquisition": "0"}, {"path": "1", "acquisition": "1"}],
            [{"path": 0}, {"path": 1}],
            [0, 1],
        ),
    )
    def test_invalid_images(self, images):
        with pytest.raises(ValueError):
            write_well_metadata(self.root, images)

    def test_unspecified_images_keys(self):
        images = [
            {"path": "0", "acquisition": 1, "unspecified_key": "alpha"},
            {"path": "1", "acquisition": 2, "unspecified_key": "beta"},
            {"path": "2", "acquisition": 3, "unspecified_key": "gamma"},
        ]
        write_well_metadata(self.root, images)
        assert "well" in self.root.attrs
        assert self.root.attrs["well"]["images"] == images
        assert self.root.attrs["well"]["version"] == FormatV04().version


class TestLabelWriter:
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.path = pathlib.Path(tmpdir.mkdir("data"))
        # create zarr v2 group...
        self.store = parse_url(self.path, mode="w", fmt=FormatV04()).store
        self.root = zarr.group(store=self.store)
        # create zarr v3 group...
        self.path_v3 = self.path / "v3"
        store_v3 = parse_url(self.path_v3, mode="w").store
        self.root_v3 = zarr.group(store=store_v3)

    def create_image_data(self, group, shape, scaler, fmt, axes, transformations):
        rng = np.random.default_rng(0)
        data = rng.poisson(10, size=shape).astype(np.uint8)
        write_image(
            image=data,
            group=group,
            scaler=scaler,
            fmt=fmt,
            axes=axes,
            coordinate_transformations=transformations,
            storage_options=dict(chunks=(128, 128)),
        )

    @pytest.fixture(
        params=(
            (1, 2, 1, 256, 256),
            (3, 512, 512),
            (256, 256),
        ),
        ids=["5D", "3D", "2D"],
    )
    def shape(self, request):
        return request.param

    @pytest.fixture(params=[True, False], ids=["scale", "noop"])
    def scaler(self, request):
        if request.param:
            return Scaler()
        else:
            return None

    def verify_label_data(
        self, img_path, label_name, label_data, fmt, shape, transformations
    ):
        # Verify image data
        reader = Reader(parse_url(f"{img_path}/labels/{label_name}"))
        node = next(iter(reader()))
        assert Multiscales.matches(node.zarr)
        if fmt.version in ("0.1", "0.2"):
            # v0.1 and v0.2 MUST be 5D
            assert node.data[0].ndim == 5
        else:
            assert node.data[0].shape == shape

        if fmt.version not in ("0.1", "0.2", "0.3"):
            for transf, expected in zip(
                node.metadata["coordinateTransformations"], transformations
            ):
                assert transf == expected
            assert len(node.metadata["coordinateTransformations"]) == len(node.data)
        assert np.allclose(label_data, node.data[0][...].compute())

        # Verify label metadata
        label_root = zarr.open(f"{img_path}/labels", mode="r")
        label_attrs = label_root.attrs
        if fmt.version == "0.5":
            label_attrs = label_attrs["ome"]
        assert "labels" in label_attrs
        assert label_name in label_attrs["labels"]

        label_group = zarr.open(f"{img_path}/labels/{label_name}", mode="r")
        imglabel_attrs = label_group.attrs
        if fmt.version == "0.5":
            imglabel_attrs = imglabel_attrs["ome"]
            assert imglabel_attrs["version"] == fmt.version
        else:
            assert imglabel_attrs["image-label"]["version"] == fmt.version
        assert "image-label" in imglabel_attrs

        # Verify multiscale metadata
        name = imglabel_attrs["multiscales"][0].get("name", "")
        assert label_name == name

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV01, id="V01"),
            pytest.param(FormatV02, id="V02"),
            pytest.param(FormatV03, id="V03"),
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_write_labels(self, shape, scaler, format_version, array_constructor):
        fmt = format_version()
        if fmt.version == "0.5":
            img_path = self.path_v3
            group = self.root_v3
        else:
            img_path = self.path
            group = self.root

        axes = "tczyx"[-len(shape) :]
        transformations = []
        for dataset_transfs in TRANSFORMATIONS:
            transf = dataset_transfs[0]
            # e.g. slice [1, 1, z, x, y] -> [z, x, y] for 3D
            transformations.append(
                [{"type": "scale", "scale": transf["scale"][-len(shape) :]}]
            )
            if scaler is None:
                break

        # create the actual label data
        label_data = np.random.randint(0, 1000, size=shape)
        if fmt.version in ("0.1", "0.2"):
            # v0.1 and v0.2 require 5d
            expand_dims = (np.s_[None],) * (5 - len(shape))
            label_data = label_data[expand_dims]
            assert label_data.ndim == 5
        label_name = "my-labels"
        label_data = array_constructor(label_data)

        # create the root level image data
        self.create_image_data(group, shape, scaler, fmt, axes, transformations)

        write_labels(
            label_data,
            group,
            scaler=scaler,
            name=label_name,
            fmt=fmt,
            axes=axes,
            coordinate_transformations=transformations,
        )
        self.verify_label_data(
            img_path, label_name, label_data, fmt, shape, transformations
        )

    @pytest.mark.parametrize(
        "format_version",
        (
            pytest.param(FormatV01, id="V01"),
            pytest.param(FormatV02, id="V02"),
            pytest.param(FormatV03, id="V03"),
            pytest.param(FormatV04, id="V04"),
            pytest.param(FormatV05, id="V05"),
        ),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_write_multiscale_labels(
        self, shape, scaler, format_version, array_constructor
    ):
        fmt = format_version()
        if fmt.version == "0.5":
            img_path = self.path_v3
            group = self.root_v3
        else:
            img_path = self.path
            group = self.root
        axes = "tczyx"[-len(shape) :]
        transformations = []
        for dataset_transfs in TRANSFORMATIONS:
            transf = dataset_transfs[0]
            # e.g. slice [1, 1, z, x, y] -> [z, x, y] for 3D
            transformations.append(
                [{"type": "scale", "scale": transf["scale"][-len(shape) :]}]
            )

        # create the actual label data
        label_data = np.random.randint(0, 1000, size=shape)
        if fmt.version in ("0.1", "0.2"):
            # v0.1 and v0.2 require 5d
            expand_dims = (np.s_[None],) * (5 - len(shape))
            label_data = label_data[expand_dims]
            assert label_data.ndim == 5
        label_data = array_constructor(label_data)

        label_name = "my-labels"
        if scaler is None:
            transformations = [transformations[0]]
            labels_mip = [label_data]
        else:
            labels_mip = scaler.nearest(label_data)

        # create the root level image data
        self.create_image_data(group, shape, scaler, fmt, axes, transformations)

        write_multiscale_labels(
            labels_mip,
            group,
            name=label_name,
            fmt=fmt,
            axes=axes,
            coordinate_transformations=transformations,
        )
        self.verify_label_data(
            img_path, label_name, label_data, fmt, shape, transformations
        )

    @pytest.mark.parametrize(
        "fmt",
        (pytest.param(FormatV04(), id="V04"), pytest.param(FormatV05(), id="V05")),
    )
    @pytest.mark.parametrize("array_constructor", [np.array, da.from_array])
    def test_two_label_images(self, array_constructor, fmt):
        if fmt.version == "0.5":
            img_path = self.path_v3
            group = self.root_v3
        else:
            img_path = self.path
            group = self.root
        axes = "tczyx"
        transformations = []
        for dataset_transfs in TRANSFORMATIONS:
            transf = dataset_transfs[0]
            transformations.append([{"type": "scale", "scale": transf["scale"]}])

        # create the root level image data
        shape = (1, 2, 1, 256, 256)
        scaler = Scaler()
        self.create_image_data(
            group,
            shape,
            scaler,
            axes=axes,
            fmt=fmt,
            transformations=transformations,
        )

        label_names = ("first_labels", "second_labels")
        for label_name in label_names:
            label_data = np.random.randint(0, 1000, size=shape)
            label_data = array_constructor(label_data)
            labels_mip = scaler.nearest(label_data)

            write_multiscale_labels(
                labels_mip,
                group,
                name=label_name,
                fmt=fmt,
                axes=axes,
                coordinate_transformations=transformations,
            )
            self.verify_label_data(
                img_path, label_name, label_data, fmt, shape, transformations
            )

        # Verify label metadata
        label_root = zarr.open(f"{img_path}/labels", mode="r")
        attrs = label_root.attrs
        if fmt.version == "0.5":
            attrs = attrs["ome"]
        assert "labels" in attrs
        assert len(attrs["labels"]) == len(label_names)
        assert all(label_name in attrs["labels"] for label_name in label_names)
